import numpy as np

##-\-\-\-\-\-\-\-\-\
## PRIVATE FUNCTIONS
##-/-/-/-/-/-/-/-/-/

# ------------------------
# Get the size of the atom
def _get_atom_size(atom_name):

    # All values in Angstrom
    all_radii = {
    'H':1.10,
    'C':1.70,
    'N':1.55,
    'O':1.52,
    'P':1.80,
    'S':1.80,
    }

    return all_radii[ atom_name ]

# -------------------------------------
# Merge all the positions of the system
def _merge_positions(system_positions):

    # Process all residues separately
    all_positions = []

    for resname in system_positions.keys():
        crt_positions = system_positions[resname]

        # Reshape the position array
        crt_positions = np.reshape(crt_positions, (crt_positions.shape[0]*crt_positions.shape[1],3))

        # Append the positions to the existing array
        all_positions.append(crt_positions)

    # Merge the arrays
    all_positions = np.concatenate( all_positions, axis=0 )

    return all_positions

##-\-\-\-\
## CLASSES
##-/-/-/-/

# --------------------------
# Class to define a molecule
class Molecule:
    def __init__(self, id, data):

        # Write the data
        self.id = id
        self.resname = data['resname']
        self.names = data['names']
        self.positions = data['positions']

        # Get the size array
        self.radii = np.array( [_get_atom_size(x) for x in self.names] )

# --------------------------------
# Class to define the whole system
class System:
    def __init__(self, molecule_data):

        # Process all the data
        self.molecules = {}
        for mol_ID in molecule_data.keys():
            self.molecules[ mol_ID ] = Molecule( mol_ID, molecule_data[ mol_ID ] )

    ##-\-\-\-\-\-\-\-\
    ## ACCESS FUNCTIONS
    ##-/-/-/-/-/-/-/-/

    # --------------------
    # Get the molecule IDs
    def getIDs(self):
        return np.array( list(self.molecules.keys()) )

    # ---------------------
    # Get all the positions
    def getPositions(self, single_array=False):

        # Process all the molecules
        all_positions = {}
        for id in self.getIDs():

            # Get the information
            crt_resname = self.molecules[id].resname
            crt_positions = self.molecules[id].positions

            # Add the resname to the dictionary if missing
            if crt_resname not in all_positions.keys():
                all_positions[crt_resname] = []

            # Add the positions to the dictionary
            all_positions[crt_resname].append( crt_positions )

        # Convert the lists to array
        for resname in all_positions.keys():
            all_positions[resname] = np.array( all_positions[resname] )

        # Return everything into a single array
        if single_array:
            return _merge_positions(all_positions)

        else:
            return all_positions

##-\-\-\-\-\-\-\-\
## PUBLIC FUNCTIONS
##-/-/-/-/-/-/-/-/

# -----------------------------------
# Load the data into the System class
def get_class(data):
    return System( data )



import stl

from md2stl.build_structure import build_atoms, scale_system
from md2stl.extract_molecules import open_file
from md2stl.system_class import get_class

##-\-\-\-\-\-\-\-\-\-\-\-\-\
## MD SIMULATION MANIPULATION
##-/-/-/-/-/-/-/-/-/-/-/-/-/

# -----------------------
# Load the structure file
def loadFile(structure_file, **kwargs):

    """Function to load a simulation file and extract the system to render.
    The list of compatible files can be found at https://userguide.mdanalysis.org/stable/formats/index.html

    INPUT(S):
    - structure_file {path-like str} : Simulation file to open.
    - hydrogens {bool} : (Opt.) Keep the hydrogen atoms. Default is False.
    - selection {str} : (Opt.) Text selection to apply on the system file. Override the other selectors listed below. Default is all.
    - resnames {str or list of str} : (Opt.) List of the residue names to keep in the representation. Default is all.
    - names {str or list of str} : (Opt.) List of the atom names to keep in the representation. Default is all.
    - resids {int or list of int} : (Opt.) List of the residue IDs to keep in the representation. Default is all.

    OUTPUT(S):
    - system_class {System class instance} : System containing all the molecules of the system and the relevant information.
    """

    # Open the simulation file to extract the informations
    data = open_file(structure_file, **kwargs)

    # Convert the data into a system class
    system_class = get_class(data)

    return system_class

##-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\
## PRODUCTION OF THE 3D ELEMENTS
##-/-/-/-/-/-/-/-/-/-/-/-/-/-/-/

# --------------------------------------------------------------
# Convert the 3d positions and atom types into a 3d construction
def build3D(system_class, **kwargs):

    """Function to build the 3D objects based on the given system.

    INPUT(S):
    - system_class {System class instance} : System containing all the molecules of the system and the relevant information. Typically extracted using the function loadFile.
    - scale {float} : (Opt.) Desired max length in millimetre of the object along the given axis. Default is 100.
    - axis {int} : (Opt.) Axis to use for the max length of the system: 0 is X, 1 is Y and 2 is Z. Default will take whatever longest length is found.
    - sphere_scale {float} : (Opt.) Additional scale factor for the size of the atom spheres. Default is 1.
    OUTPUT(S):
    - all_atoms {Mesh class instance} : Collection of meshes to be exported.
    """

    # Rescale the system to the desired size
    new_class, system_factor = scale_system(system_class, **kwargs)

    # Build the particles
    all_atoms = build_atoms(new_class, system_factor, **kwargs)

    return all_atoms

# --------------------------------
# Save the 3d meshes in a stl file
def save3D(filename, system_3d):

    """Function to save the 3D objects in a .stl file.

    INPUT(S):
    - filename {path-like str} : Path and name of the file to be saved. Extension should be .stl.
    - system_3d {Mesh class instance} : Collection of meshes to be exported.
    """

    # Save in file
    system_3d.save(filename, mode=stl.Mode.ASCII)



from copy import deepcopy
import numpy as np
from stl import mesh
from tqdm import tqdm

from md2stl.make_objects import generate_sphere

##-\-\-\-\-\-\-\-\-\
## PRIVATE FUNCTIONS
##-/-/-/-/-/-/-/-/-/

# ----------------------------
# Get the limits of the system
def _get_system_sizes(system):

    # Extract the min and max position of the system in all axis
    all_positions = system.getPositions(single_array=True)

    # Get the min and max values for each axis
    xlength = np.amax(all_positions[:,0]) - np.amin(all_positions[:,0])
    ylength = np.amax(all_positions[:,1]) - np.amin(all_positions[:,1])
    zlength = np.amax(all_positions[:,2]) - np.amin(all_positions[:,2])

    return [xlength, ylength, zlength]

# --------------------------------------
# Rescale the object by the given factor
def _rescale_objet(object, factor):

    object.x *= factor
    object.y *= factor
    object.z *= factor

    return object

# -------------------------------------
# Move the object to the given location
def _move_objet(object, new_position):

    object.x += new_position[0]
    object.y += new_position[1]
    object.z += new_position[2]

    return object

# -----------------------------------------------
# Rescale all the system to fit in the given unit
def _rescale_system(system, f):

    # Make a copy of the system
    new_system = deepcopy(system)

    # Process all the molecules
    for mol_ID in new_system.molecules.keys():
        new_system.molecules[mol_ID].positions = new_system.molecules[mol_ID].positions * f

    return new_system

# --------------------------------------------------
# Add the atom at the given position after rescaling
def _add_atom(sphere, positions, radius):

    # Copy the unit sphere
    atom = mesh.Mesh(sphere.data.copy())

    # Rescale the sphere
    atom = _rescale_objet(atom, radius)

    # Move the sphere
    atom = _move_objet(atom, positions)

    return atom

# ---------------------------------
# Populate the space with the atoms
def _populate_atoms(system, sphere):

    # Initialise the 3d space
    all_atoms = []

    # Loop over all the molecules
    for mol_ID in tqdm(system.molecules.keys()):

        # Get the current atom and positions
        crt_positions = system.molecules[mol_ID].positions
        crt_radii = system.molecules[mol_ID].radii

        # Loop over all the atoms
        for atom_ID in range(crt_positions.shape[0]):

            # Get the new object
            new_atom = _add_atom( sphere, crt_positions[atom_ID], crt_radii[atom_ID])

            # Append the new object
            all_atoms.append( new_atom.data.copy() )

    # Merge the meshes together
    all_atoms = mesh.Mesh(np.concatenate(all_atoms))

    return all_atoms

##-\-\-\-\-\-\-\-\
## PUBLIC FUNCTIONS
##-/-/-/-/-/-/-/-/

# ------------------
# Rescale the system
def scale_system(system, **kwargs):

    # Get the arguments
    scale = kwargs.get('scale', 100)
    axis = kwargs.get('axis', None)

    # Get the limits of the system
    max_lengths = _get_system_sizes(system)

    # Select the length to scale by
    if axis is None:
        length = np.amax(max_lengths)
    else:
        length = max_lengths[axis]

    # Calculate the scaling factor
    f = scale / length

    # Rescale the positions
    new_system = _rescale_system(system, f)

    return new_system, f

# ------------------------------
# Build the atoms for the system
def build_atoms(system, system_factor, **kwargs):

    # Get the arguments
    sphere_scale = kwargs.get('sphere_scale', 1)

    # Get the unit sphere
    unit_sphere = generate_sphere(**kwargs)

    # Rescale the base sphere
    total_factor =  sphere_scale * system_factor * .7 # .7 is an arbitrary normalisation factor
    unit_sphere = _rescale_objet(unit_sphere, total_factor)

    # Fill the space with the atoms
    all_atoms = _populate_atoms(system, unit_sphere)

    return all_atoms



import MDAnalysis as md
import numpy as np

##-\-\-\-\-\-\-\-\-\
## PRIVATE FUNCTIONS
##-/-/-/-/-/-/-/-/-/

# ------------------------------------
# Turn the input selection into a text
def _get_selection(selection, selection_type = 'resname'):

    # Convert the input if it is not a list or tuple
    if isinstance(selection, str) or isinstance(selection, int):
        selection = [selection]

    # Convert into string
    # Single element
    if len(selection) == 1:
        selection_text = selection_type + ' ' + str( selection[0] )

    # List of several elements
    else:
        selection_text = '(' + selection_type + ' ' + str(selection[0])

        for item in selection[1:]:
            selection_text += ' or ' + selection_type + ' ' + str(item)

        selection_text += ')'

    return selection_text

# --------------------------------------------
# Select the molecules to extract specifically
def _select_molecules(system, **kwargs):

    # Read the argument
    add_hydrogen = kwargs.get('hydrogens', True)
    text_selection = kwargs.get('selection', None)
    resname_selection = kwargs.get('resnames', None)
    name_selection = kwargs.get('names', None)
    resid_selection = kwargs.get('resids', None)

    # Check if a text selection has been given
    if text_selection is not None:
        selection = system.select_atoms(text_selection)

    # Check if any other type of selection is given
    elif resname_selection is not None or name_selection is not None or resid_selection is not None:

        # Initialise the text selection
        text_selection = ''

        # Add residue name selection
        if resname_selection is not None:
            text_selection += _get_selection(resname_selection, selection_type='resname')

        # Add atom name selection
        if name_selection is not None:

            # Add separator if needed
            if text_selection != '':
                text_selection += ' and '

            # Add the text
            text_selection += _get_selection(name_selection, selection_type='name')

        # Add the residue id selection
        if resid_selection is not None:

            # Add separator if needed
            if text_selection != '':
                text_selection += ' and '

            # Add the text
            text_selection += _get_selection(resid_selection, selection_type='resid')

        # Remove the hydrogen if requested
        if not add_hydrogen:
            text_selection += ' and not type H'

        # Get the selection
        selection = system.select_atoms(text_selection)

    # Select everything otherwise
    else:

        # Remove the hydrogen if requested
        if not add_hydrogen:
            text_selection = 'not type H'
        else:
            text_selection = 'all'

        selection = system.select_atoms(text_selection)

    return selection

# ---------------------------------------------------
# Read the given selection to extract the information
def _read_selection(selection):

    # Get the list of all residue ids
    list_resids = np.unique( selection.resids )

    # Loop over all the IDs
    all_data = {}
    for id in list_resids:

        # Extract the residue name
        crt_resname = selection.resnames[ selection.resids == id ][0]

        # Extract the atom names
        crt_names = np.copy( selection.types[ selection.resids == id ] )

        # Extract the atom positions
        crt_positions = np.copy( selection.positions[ selection.resids == id ] )

        # Add to the dictionary
        all_data[ id ] = {
        'resname' : crt_resname,
        'names' : crt_names,
        'positions' : crt_positions
        }

    return all_data

##-\-\-\-\-\-\-\-\
## PUBLIC FUNCTIONS
##-/-/-/-/-/-/-/-/

# ----------------
# Open the MD file
def open_file(structure_file, **kwargs):

    # Load the system from the file
    system = md.Universe(structure_file)

    # Refine the selection if required
    selection = _select_molecules(system, **kwargs)

    # Extract the information from the selection
    data = _read_selection(selection)

    return data



import numpy as np
from stl import mesh

##-\-\-\-\-\-\
## BASE FIGURES
##-/-/-/-/-/-/

"""Code for sphere creation taken from:
http://blog.andreaskahler.com/2009/06/creating-icosphere-mesh-in-code.html
All credits go to the author"""

# -----------------------------------
# Get the vertices of the icosahedron
def _get_icosahedron_vertices():

    # Get the base variable
    t = (1 + np.sqrt(5)) / 2

    # Define the vertices
    vertices = np.array([
    [-1, t, 0],
    [1, t, 0],
    [-1, -t, 0],
    [1, -t, 0],

    [0, -1, t],
    [0, 1, t],
    [0, -1, -t],
    [0, 1, -t],

    [t, 0, -1],
    [t, 0, 1],
    [-t, 0, -1],
    [-t, 0, 1]
    ])

    # Normalise by length
    length = np.sqrt( np.sum(vertices**2, axis=1) )
    vertices = vertices / length[:,np.newaxis]

    return vertices

# --------------------------------
# Get the faces of the icosahedron
def _get_icosahedron_faces():

    # Define the faces
    faces = np.array([
    [0,11,5],
    [0,5,1],
    [0,1,7],
    [0,7,10],
    [0,10,11],

    [1,5,9],
    [5,11,4],
    [11,10,2],
    [10,7,6],
    [7,1,8],

    [3,9,4],
    [3,4,2],
    [3,2,6],
    [3,6,8],
    [3,8,9],

    [4,9,5],
    [2,4,11],
    [6,2,10],
    [8,6,7],
    [9,8,1]
    ])

    return faces

##-\-\-\-\-\-\-\-\-\
## PRIVATE FUNCTIONS
##-/-/-/-/-/-/-/-/-/

# ------------------------------------------------------------
# Get the middle point of the two triangles on the unit sphere
def _get_middle_point(vertices, a, b):

    # Sort the indices
    a, b = np.sort([a, b])

    # Get the points
    pa, pb = vertices[a], vertices[b]

    # Calculate the middle point
    mid_p = (pa + pb)/2

    # Correct by length
    len_p = np.sqrt( np.sum(mid_p**2) )
    mid_p = mid_p / len_p

    # Add the new point to the vertices
    new_index = len(vertices)
    vertices = np.vstack([vertices, mid_p])

    return new_index, vertices

# --------------------
# Refine the triangles
def _refine_sphere_triangles(vertices, faces):

    # Loop over all the triangles
    for vx,vy,vz in faces:

        # Define new points for the new triangles
        a, vertices = _get_middle_point(vertices, vx, vy)
        b, vertices = _get_middle_point(vertices, vy, vz)
        c, vertices = _get_middle_point(vertices, vz, vx)

        # Add the new face
        faces = np.vstack( [faces,
        np.array([vx, a, c]),
        np.array([vy, b, a]),
        np.array([vz, c, b]),
        np.array([a, b, c])
        ] )

    return vertices, faces

##-\-\-\-\-\-\-\-\
## PUBLIC FUNCTIONS
##-/-/-/-/-/-/-/-/

# -------------------
# Generate the sphere
def generate_sphere(**kwargs):

    # Extract the parameters
    resolution = kwargs.get('resolution',3)

    # Get the sphere vertices and faces
    vertices = _get_icosahedron_vertices()
    faces = _get_icosahedron_faces()

    # Refine the mesh
    if resolution > 0:

        # Repeat as much as needed
        for i in range(resolution):
            vertices, faces = _refine_sphere_triangles(vertices, faces)

    # Create the mesh
    sphere = mesh.Mesh(np.zeros(faces.shape[0], dtype=mesh.Mesh.dtype))
    for i, f in enumerate(faces):
        for j in range(3):
            sphere.vectors[i][j] = vertices[f[j],:]

    return sphere



from setuptools import setup

setup(
	name = "md2stl",
	version = "0.1",
	author = "Vivien WALTER",
	author_email = "walter.vivien@gmail.com",
	description = (
	"Python tools to convert a MD simulation file into a .stl file ready to be 3d printed."
	),
	license = "GPL3.0",
	url='https://github.com/vivien-walter/md2stl',
	download_url = 'https://github.com/vivien-walter/md2stl/archive/v0.1.tar.gz',
	packages=[
	'md2stl'
	],
	install_requires=[
	'numpy',
	'numpy-stl',
	'MDAnalysis',
	'tqdm'
	]
)
